package Warsync;

##############################################################################
## Warsync.pm - Warsync                                                     ##
##                                                                          ##
## Copyright (C) 2001-2004 Paul J. Baker                                    ##
##                                                                          ##
## This program is free software; you can redistribute it and/or modify     ##
## it under the terms of the GNU General Public License version 2 as        ##
## published by the Free Software Foundation.                               ##
##                                                                          ##
## This program is distributed in the hope that it will be useful,          ##
## but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            ##
## GNU General Public License for more details.                             ##
##                                                                          ##
## You should have received a copy of the GNU General Public License        ##
## along with this program; if not, write to the Free Software              ##
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                ##
##                                                                          ##
## You may contact the author via electronic mail at pfars@paulbaker.net    ##
##############################################################################

## $Id: Warsync.pm,v 1.17 2004/02/04 10:31:45 pbaker Exp $

=head1 NAME

Warsync - Wrapper Around Rsync

=head1 SYNOPSIS

  use Warsync;

=head1 DESCRIPTION

Warsync exports various functions used by B<Warsync> scripts.

=head1 AUTHOR

Paul Baker, pbaker@where2getit.com

=head1 SEE ALSO

warsync(8).

=head1 VERSION

0.9.6

=head1 FUNCTIONS

=cut

#############################################################################

require 5.6.0;
use strict;
use warnings;

use Carp;
use Getopt::Long ();
use Pod::Usage;
use Sys::Hostname;

require Exporter;

our @ISA = qw( Exporter );

our @EXPORT = qw( error is_ipaddr is_fqdn );

our $VERSION = '0.9.6';

## set option defaults.
Getopt::Long::Configure("bundling");

sub options
{
    my ($class, @args) = @_;

    my %option = ();
    
    Getopt::Long::GetOptions(
        \%option,
        'help|H|?',
        'man',
        'version|V',
        @args
    ) or pod2usage(2);

    pod2usage(1) if $option{help};
    pod2usage(-exitstatus => 0, -verbose => 2) if $option{man};

    # display version number if --version set.
    if ($option{version}) {
        print "Warsync - Wrapper Around Rsync\n";
        print "Version $VERSION\n";
        exit(0);
    }
    
    return \%option;
}

sub error ($;$) {
    print STDERR 'Error: ', $_[0], "\n";
    exit ($_[1] || 1);
}

sub prompt
{
    my ($class, %args) = @_;
    
    my $response = '';
    my @choices;
    
    if (exists $args{default}) {
        $response = $args{default};
    }
    elsif (exists $args{choices}) {
        $response = $args{choices}[0];
        @choices  = @{$args{choices}};
    }
    
    for (;;) {
        ## print the question.
        print $args{prompt};
        
        ## print list of choices.
        if (@choices) {
            print '(';
            print join( '/', map { (lc($response) eq lc($_)) 
                                   ? uc($_) : $_ } @choices );
            print ') ';
        }
        ## or print the default.
        elsif ($response) {
            print '[', $response, '] ';
        }
        ## mark end of prompt.
        print '>> ';
        ## get the response.
        chomp(my $in = <STDIN>);
        ## use if not blank.
        $response = $in if length($in);
        
        ## check that it matches a choice
        if (@choices) {
            foreach my $choice (@choices) {
                ## return if it matches.
                return lc($response) if lc($response) eq lc($choice);
            }
        }
        ## no specific choices, return any response.
        elsif ($response) {
            return $response;
        }
    }
}

sub is_ipaddr ($) {
    my $in = shift;
    
    my ($ip) = $in =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/;
    
    return $ip;
}


{
    ## hostname/domain regex
    my $rfc1035_label_re  = qr/[0-9A-Za-z](?:[-0-9A-Za-z]*[0-9A-Za-z])?/;

    sub is_fqdn ($) {
        my $in = shift || '';
        
        my ($hostname) = $in =~ /($rfc1035_label_re(?:\.$rfc1035_label_re)*)/;
        
        return $hostname;
    }
}

sub deep_copy ($);
sub deep_copy ($)
{
    ## http://www.stonehenge.com/merlyn/UnixReview/col30.html
    my $this = shift;
    if (not ref $this) {
        $this;
    }
    elsif (ref $this eq "ARRAY") {
        [map deep_copy($_), @$this];
    }
    elsif (ref $this eq "HASH") {
        +{map { $_ => deep_copy($this->{$_}) } keys %$this};
    }
    else {
        die "what type is $_?"
    }
}

sub parse_package_list ($)
{
    my $dpkg = shift;
    my %packages = ();
        
    while (my $line = <$dpkg>) {
        ## skip header lines.
        next if $line =~ /^[|+]|Desired/;
        chomp($line);
        
        if (my @pkg = $line =~ 
                /^
                (...) \s+              ## 0 - desired state
                ([-a-z0-9+.]+) \s+     ## 1 - package
                ([-.0-9+a-zA-Z:]+) \s+ ## 2 - version
                (\S.*)                 ## 3 - description
                /x)
        {
            $packages{$pkg[1]} = \@pkg;
        }
    }
    
    return \%packages;
}

sub real_package_version ($)
{
    my $upkg = shift; ## debian package name
    my $ver;          ## version to be.

    ## untaint package name.
    my ($pkg) = $upkg =~ /([-a-z0-9+.]+)/;
    
    ## read package information.
    my @cmd = ( 'dpkg', '-s', $pkg ); 
    open(my $exec_h, '-|') || exec( @cmd ) 
        or die "error '$!' trying to exec: @cmd"; 
    
    while (<$exec_h>) {
        if (($ver) = $_ =~ /^Version: ([-.0-9+a-zA-Z:]+)/) {
            last;
        }
    }
    
    close($exec_h); 

    return $ver if $ver;
    
    croak "could not find version for $pkg";
}

#############################################################################
#############################################################################

package Warsync::Server;

sub new
{
    my $self = shift;
    my $class = ref($self) || $self;
    return $self if ref $self;

    unless (-r '/etc/warsync/server-key' and -r '/etc/warsync/client.list') {
        return undef;
    }

    ## read in the client list.
    open(my $client_list, '/etc/warsync/client.list')
        or die "could not open client list: $!";
    
    my %clients = ();
    while (my $line = <$client_list>) {
        my ($client, $fqdn) = split ':', $line, 2;
        $clients{$client} = Warsync::is_fqdn($fqdn);
    }
    
    close($client_list);
    
    return bless { clients => \%clients } => $class;
}

sub apt_conf
{
    my $self = shift->new;
    
    return '' unless -e '/etc/apt/apt.conf';
    
    open(my $fh, '/etc/apt/apt.conf')
        or Warsync::error "could not read /etc/apt/apt.conf";
    my $slurp;
    { local $/ = undef; $slurp = <$fh> }
    close($fh);
    
    return $slurp;
}

sub apt_preferences
{
    my $self = shift->new;
    
    return '' unless -e '/etc/apt/preferences';
    
    open(my $fh, '/etc/apt/preferences')
        or Warsync::error "could not read /etc/apt/preferences";
    my $slurp;
    { local $/ = undef; $slurp = <$fh> }
    close($fh);
    
    return $slurp;
}

sub apt_sources_list
{
    my $self = shift->new;
    
    open(my $fh, '/etc/apt/sources.list')
        or Warsync::error "could not read /etc/apt/sources.list";
    my $slurp;
    { local $/ = undef; $slurp = <$fh> }
    close($fh);
    
    return $slurp;
}

sub client_command
{
    my $self   = shift->new;
    my $client = shift;
    my @cmd    = @_;
    
    my $fqdn = $self->client_fqdn($client) || Warsync::is_fqdn($client);

    my @scmd = ( 'ssh', '-o', 'NumberOfPasswordPrompts=0',
                        '-i', '/etc/warsync/server-key',
                        '-T', $fqdn, @cmd );
    
    open(my $ssh_h, '-|') || exec( @scmd )
        or die "error '$!' trying to exec: @scmd";
    
    return $ssh_h;
}

sub client_gui_command
{
    my $self   = shift->new;
    my $client = shift;
    my @cmd    = @_;
    
    my $fqdn = $self->client_fqdn($client) || Warsync::is_fqdn($client);
    
    my @scmd = ( 'ssh', '-o', 'NumberOfPasswordPrompts=0',
                        '-i', '/etc/warsync/server-key',
                        '-qt', $fqdn, @cmd );
    
    system( @scmd ) and return undef;
    
    return 1;
}

sub client_fqdn
{
    my $self   = shift->new;
    my $client = shift;
    
    if (exists $self->{clients}{$client}) {
        return $self->{clients}{$client};
    }
    
    return undef;
}

sub client_replicate_debian
{
    my $self   = shift->new;
    my $client = shift; ## client to replicate
    my $option = shift; ## options
    
    print "Debian Package Syncronization\n";
    print "* checking client package installations...";

    my $cmd_h = $self->client_command($client, 'warsync-echo-package-list');
    my $spkgs = $self->packages;
    my $cpkgs = Warsync::parse_package_list($cmd_h);
    close($cmd_h);
    
    my @install = ();
    my @remove  = ();
    my @purge   = ();

    ## compare server packages to the client and create
    ## list of apt-get commands to run on the client.
    foreach my $pkg (keys %$spkgs) {

        my $spkg = $spkgs->{$pkg};
       
        #print STDERR $spkg->[0];
 
        ## if installed, make sure client installs it.
        if ($spkg->[0] eq 'ii ') {
            ## client has package installed
            if (exists $cpkgs->{$pkg} and $cpkgs->{$pkg}[0] eq 'ii ') {
                ## check version.
                if ($spkg->[2] ne $cpkgs->{$pkg}[2]) {
                    ## need to upgrade/downgrade version. determine
                    ## distribution.
                    #print STDERR "server has package $pkg $spkg->[2]. client has $cpkgs->{$pkg}[2].\n";
                    my $ver = Warsync::real_package_version($pkg);
                    push @install, "$pkg=$ver";
                }
            }
            ## client does not have package installed.
            else {
                #print STDERR "client missing package $pkg $spkg->[2].\n";
                my $ver = Warsync::real_package_version($pkg);
                push @install, "$pkg=$ver";
            }       
        }
        
        ## if removed, make sure client removes it.
        elsif ($spkg->[0] =~ /^[rp]c $/) {
            ## client has package installed
            if (exists $cpkgs->{$pkg} and $cpkgs->{$pkg}[0] !~ /^[rp]c $/) {
                #print STDERR "server has removed package $pkg. client has installed.\n";
                push @remove, $pkg;
            }
        }
        
        ## unexpected state, possibly flagged for removal,
        ## but not yet uninstalled.
        else {
            print STDERR "warn: server package $pkg $spkg->[2] has unexpected state: $spkg->[0].\n";
        }

        ## mark package as processed.
        $spkg->[5] = $cpkgs->{$pkg}[5] = 1;

    }
                
    ## now compare remaining client packages to the server.
    foreach my $pkg (keys %$cpkgs) {
        ## skip if we already processed it.
        next if $cpkgs->{$pkg}[5];
        #print STDERR "client has package $pkg installed. server does not.\n";
        ## it shouldn't be here if it's not on the server.
        push @purge, $pkg;
    }

    ## return if there is nothing to do.
    unless (@install || @remove || @purge) {
        print "client up-to-date.\n";
        return;
    }

    #print STDERR "Install: @install\n";
    #print STDERR "Remove:  @remove\n";
    #print STDERR "Purge:   @purge\n";
   
    ## have client update package sources.
    print "client out-of-sync.\n";
    print "* syncronizing package sources...";
    
    my $dryrun = $option->{n} ? '-dry-run' : '';
    
    ## run as gui command if TTY exists.
    if (-t) {
        print "\n";
        $self->client_gui_command(
                $client, "warsync-sync-package-sources$dryrun" )
            or Warsync::error "client warsync-package-source-sync returned status: $?";
    }
    ## otherwise run quietly.
    else {
        my $comh = $self->client_command(
                $client, "warsync-sync-package-sources$dryrun" );
        print <$comh>;
        close($comh)
            or Warsync::error "client warsync-package-source-sync returned status: $?";
        print "done.\n";
    }

    print "* installing/removing packages on client with apt-get...\n";
    
    my @apt = "warsync-apt-get-install$dryrun";
    push @apt, @install; ## packages to install
    push @apt, map { "$_-" } @remove;              ## packages to remove
    push @apt, map { "$_-" } @purge;               ## packages to purge
    
    if (-t) {
        $self->client_gui_command($client, @apt)
            or Warsync::error "client warsync-apt-get-install returned status: $?";
    }
    else {
        my $comh = $self->client_command($client, @apt);
        print <$comh>;
        close($comh)
            or Warsync::error "client warsync-apt-get-install returned status: $?";
    }

    if (@purge) {
        print "* purging packages on client with apt-get...\n";
        my @apt = ("warsync-apt-get-purge$dryrun", @purge);
        if (-t) {
            $self->client_gui_command($client, @apt)
                or Warsync::error "client warsync-apt-get-purge returned status: $?";
        }
        else {
            my $comh = $self->client_command($client, @apt);
            print <$comh>;
            close($comh)
                or Warsync::error "client warsync-apt-get-purge returned status: $?";
        }
    }
        
    print "* package syncronization completed!\n";
}

sub clients
{
    my $self = shift->new;
    
    return keys %{ $self->{clients} };
}

sub name
{
    my $self = shift->new;
    
    return Sys::Hostname::hostname;
}

sub packages
{
    my $self = shift->new;
    
    unless ($self->{_packages}) {
    
        ## get previous COLUMNS environment variable if
        ## one existed.
        my $prev_col = $ENV{COLUMNS};
        ## now set it to 200
        $ENV{COLUMNS} = 200;
        
        ## read dpkg -l
        open(my $dpkg, 'dpkg -l |')
            or die "could not exec dpkg -l: $!";
        
        $self->{_packages} = Warsync::parse_package_list($dpkg);
        
        close($dpkg);
        
        if ($prev_col) {
            $ENV{COLUMNS} = $prev_col; 
        } 
        else { 
            delete $ENV{COLUMNS};
        }
        
    }

    return Warsync::deep_copy($self->{_packages});
}
   
sub run
{
    my $self = shift->new;
    
    my $option = Warsync->options(
        'v|verbose+',
        'q|quiet',
        'n|dry-run',
        'a|all-files',
        'client=s@',
        'debian',
    );
    
    unless ($option->{client}) {
        $option->{client} = [ $self->clients ];
    }
    
    my @filepacks = @ARGV;
    
    if ($option->{a}) {
        @filepacks = $self->filepacks;
    }
    
    ## they need to specify something to do.
    unless (@filepacks or $option->{debian}) {
        Pod::Usage::pod2usage(1);
    }
        
    ## iterate through each client and replicate.
    foreach my $client (@{ $option->{client} }) {

        print "######################################################\n";
        
        ## grab the fqdn for this client.
        my $fqdn = $self->client_fqdn($client);
        
        ## if we didn't get back a fqdn, this is a bad client name.
        unless ($fqdn) {
            print STDERR $client, " is not a valid client\n";
            next;
        }
        
        print "Replicating $client...\n";
        
        if ($option->{debian}) {
            $self->client_replicate_debian($client, $option);
        }
        
        foreach my $filepack (@filepacks) {
            $self->client_replicate_filepack($client, $filepack, $option);
        }
    }
}


#############################################################################
#############################################################################

package Warsync::Client;

sub new
{
    my $self = shift;
    my $class = ref($self) || $self;
    return $self if ref $self;

    ## only return object if this is a client system.
    ## check that we are a client by looking for the client
    ## authentication key and config file.
    unless (-r '/etc/warsync/client-key' and -r '/etc/warsync/client.conf') {
        return undef;
    }

    ## okay now parse the client config file.
    open(my $conf_fh, '/etc/warsync/client.conf')
        or Warsync::error "could not open client config file: $!";

    $self = bless {} => $class;

    while (<$conf_fh>) {
        chomp;                  # no newline
        s/#.*//;                # no comments
        s/^\s+//;               # no leading white
        s/\s+$//;               # no trailing white
        next unless length;     # anything left?
        my ($var, $value) = split(/\s*=\s*/, $_, 2);
        $self->{ lc($var) } = $value;
    }

    close($conf_fh);
    
    ## must have a server.
    unless($self->{server} = Warsync::is_fqdn($self->{server})) {
        Warsync::error "no server configured.";
    }

    return $self;
}

sub apt_get_install
{
    my $self     = shift->new;
    my $option   = shift;      ## options hash
    my $packages = shift;      ## space separated list of packages
    
    my @packages = split / /, $packages;
    
    my @apt = 'apt-get';
    push @apt, '--dry-run' if $option->{dryrun};
    
    if (-t) {
        system(@apt, 'install', @packages)
            and Warsync::error "something went wrong running apt-get install: $?";
    }
    else {
        system(@apt, '-q', '-y', 'install', @packages)
            and Warsync::error "something went wrong running apt-get install: $?";
    }
}

sub apt_get_purge
{
    my $self     = shift->new;
    my $option   = shift;      ## options hash
    my $packages = shift;      ## space separated list of packages
    
    my @packages = split / /, $packages;
    
    my @apt = 'apt-get';
    push @apt, '--dry-run' if $option->{dryrun};
    
    if (-t) {
        system(@apt, '--purge', 'remove', @packages)
            and Warsync::error "something went wrong running apt-get install: $?";
    }
    else {
        system(@apt, '-q', '-y', '--purge', 'remove', @packages)
            and Warsync::error "something went wrong running apt-get install: $?";
    }
}

sub compression
{
    my $self = shift->new;
    
    return $self->{compression} eq 'y' ? ('-C') : ();
}

sub name
{
    my $self = shift->new;
    
    return Sys::Hostname::hostname;
}

sub package_list
{
    my $self = shift->new;
    
    my $prev_col = $ENV{COLUMNS};
    $ENV{COLUMNS} = 200;
    
    open(my $dpkg, 'dpkg -l |')
        or die "could not exec dpkg -l: $!";
    
    my $list; 
    { local $/ = undef; $list = <$dpkg> } 
    close($dpkg);
    
    if ($prev_col) {
        $ENV{COLUMNS} = $prev_col;
    }
    else {
        delete $ENV{COLUMNS};
    }
    
    return $list;
} 

sub public_key
{
    my $self = shift->new;
    
    return `cat /etc/warsync/client-key.pub`;
}

sub server
{
    my $self = shift->new;
    
    return $self->{server};
}

sub server_command
{
    my $self = shift->new;
    my @cmd  = @_;
    
    my @scmd = ( 'ssh', $self->compression,
                        '-o', 'NumberOfPasswordPrompts=0',
                        '-i', '/etc/warsync/client-key',
                        $self->server, @cmd );
    
    open(my $ssh_h, '-|') || exec( @scmd )
        or die "error '$!' trying to exec: @scmd";
    
    return $ssh_h;
}

sub server_name
{
    my $self = shift->new;
    
    my $server_h = $self->server_command('warsync-echo-name');
        
    my @scom = <$server_h>;
    
    close($server_h);
    
    if ($scom[0]) {
        my ($name) = $scom[0] =~ /^name: (\S+)/;
        return $name if $name;
    }
    
    Warsync::error "could not retrieve server name.";
}

sub sync_package_sources
{
    my $self   = shift->new;
    my $option = shift;
    
    ## sync apt config.    
    {
        my $sch = $self->server_command('warsync-echo-apt-conf');
        my @data = <$sch>;
    
        close($sch) or Warsync::error "could not retrieve apt.conf: $?";

        if ($option->{dryrun}) {
            print STDERR "dry-run: syncronize /etc/apt/apt.conf\n";
        }
        else {
            open(my $fh, '> /etc/apt/apt.conf')
                or Warsync::error "could not write apt.conf: $!";
            print $fh @data;
            close($fh);
        }
    }
    
    ## sync apt preferences.
    {
        my $sch = $self->server_command('warsync-echo-apt-preferences');
        my @data = <$sch>;
    
        close($sch) or Warsync::error "could not retrieve preferences: $?";
    
        if ($option->{dryrun}) {
            print STDERR "dry-run: syncronize /etc/apt/preferences\n";
        }
        else {
            open(my $fh, '> /etc/apt/preferences')
                or Warsync::error "could not write preferences: $!";
            print $fh @data;
            close($fh);
        }
    }
    
    ## sync apt sources.list.
    {
        my $sch = $self->server_command('warsync-echo-apt-sources-list');
        my @data = <$sch>;
    
        close($sch) or Warsync::error "could not retrieve sources.list: $?";
    
        if ($option->{dryrun}) {
            print STDERR "dry-run: syncronize /etc/apt/sources.list\n";
        }
        else {
            open(my $fh, '> /etc/apt/sources.list')
                or Warsync::error "could not write to sources.list: $!";
            print $fh @data;
            close($fh);
        }
    }
    
    ## apt-get update
    {
        if ($option->{dryrun}) {
            print STDERR "dry-run: execute apt-get update\n";
        }
        elsif (-t) {
            system('apt-get', 'update')
                and Warsync::error "something went wrong running apt-get update: $?";
        }
        else {
            system('apt-get', '-q', 'update')
                and Warsync::error "something went wrong running apt-get update: $?";
        }
    }
    return 1;
}
    
1;
