package Warsync::PID;

##############################################################################
## PID.pm - Warsync PID Management Perl Module                              ##
##                                                                          ##
## Copyright (C) 2001-2004 Paul J. Baker                                    ##
##                                                                          ##
## This program is free software; you can redistribute it and/or modify     ##
## it under the terms of the GNU General Public License version 2 as        ##
## published by the Free Software Foundation.                               ##
##                                                                          ##
## This program is distributed in the hope that it will be useful,          ##
## but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            ##
## GNU General Public License for more details.                             ##
##                                                                          ##
## You should have received a copy of the GNU General Public License        ##
## along with this program; if not, write to the Free Software              ##
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                ##
##                                                                          ##
## You may contact the author via electronic mail at pfars@paulbaker.net    ##
##############################################################################

## $Id: PID.pm,v 1.1.1.1 2004/01/30 20:47:12 pbaker Exp $

=head1 NAME

Warsync::PID - Module for managing Process PID files.

=head1 SYNOPSIS

  use Warsync::PID;

  my $pid = Warsync::PID->new(file => '/path/to/pidfile');

  if ($pid->lock) {
    .. do something ..
  } else {
    die 'could not get lock: ' . $pid->error;
  }

=head1 DESCRIPTION

Warsync::PID creates a PID file for the process you are running
when B<$pid-E<gt>lock> method is called. PID file is removed when B<$pid>
goes out of scope, or when B<$pid-E<gt>unlock> is called.

=head1 EXPORT

None by default.

=head1 AUTHOR

Paul Baker, pbaker@where2getit.com

=head1 SEE ALSO

perl(1).

=head1 VERSION

1.1

=head1 FUNCTIONS

=cut

#############################################################################

require 5.6.0;

use strict;
use warnings;

use Carp;
use IO::File;
use Errno qw(ESRCH EPERM ENOENT);

use vars qw( @ISA %EXPORT_TAGS @EXPORT_OK @EXPORT $VERSION );

require Exporter;

@ISA = qw( Exporter );

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
%EXPORT_TAGS = ();

@EXPORT_OK = ();

@EXPORT = ();

our $VERSION = do { my @r = (q$Revision: 1.1.1.1 $ =~ /\d+/g); sprintf "%d." . "%02d" x $#r, @r };


=head2 new

Initializes object. Does not create a PID file.

=cut

sub new
{

    my ($class, %args) = @_;

    my %obj = %args;

    croak 'must specify file argument' unless $args{file};

    return bless \%obj, ref($class) || $class;

}

=head2 lock

Lock PID file. Return true if PID file created successfully.
Return false otherwise.

=cut

sub lock
{
    my $self = shift;
    my $step = shift || 0;

    if ($step > 10) {
        confess 'something horribly wrong is going on: ', $self->{error};
    }

    my $file = $self->{file};

    ## open the file for write only, fail if it already exists.
    my $fh = IO::File->new($file, O_WRONLY|O_EXCL|O_CREAT);

    ## opened file, hurray!
    if ($fh) {

        ## write the current pid to the file.
        print $fh "$$\n";
        ## make sure it really gets written.
        $fh->flush;
        ## now hold onto this file handle for later.
        $self->{_fh} = $fh;

        ## return success
        return 1;
    }
    ## open failed, check if the file exists then.
    else {

        ## lets hold unto the error from the first open just
        ## in case.
        $self->{error} = $!;
        $self->{bang}  = $!;

        ## try and open it for read.
        $fh = IO::File->new($file, O_RDONLY);

        ## if we still couldn't open, let's die
        return undef unless $fh;

        chomp(my $pid = <$fh>);

        $fh->close;

        ## check that we got a pid number from the file.
        if ($pid) {

            ## good, let's check if that pid is running.
            if (kill 0, $pid) {
                $self->{error} = 'process already running';
                carp $self->{error} if $^W;
                $self->{pid}   = $pid;
                return 0;
            }
            ## may not have been able to send signal because process
            ## is outside of our control.
            elsif ($! == EPERM) {
                $self->{error} = 'process already running';
                carp $self->{error} if $^W;
                $self->{bang}  = $!;
                $self->{pid}   = $pid;
                return 0;
            }
            ## couldn't send signal because process is dead.
            elsif ($! == ESRCH) {
                $self->{error} = "found PID $pid but it appears to be dead.";
                carp $self->{error} if $^W;
                ## remove the file so we can try again.
                if (unlink $file or $! == ENOENT) {
                    ## removed file successfully. try locking again.
                    return $self->lock(++$step);
                }
                ## couldn't delete file?
                else {
                    $self->{error} = "unable to remove invalid pid file: $!";
                    carp $self->{error} if $^W;
                    return undef;
                }
            }
        }
        ## that's strange file is there but no pid is in it.
        ## perhaps we've already been through this though?
        elsif ($step) {

            ## let's just skip out.
            $self->{error} = 'found emtpy pid file';
            carp $self->{error} if $^W;
            return undef;

        }
        ## but if this is the first time...
        else {
            ## let's wait for a bit and try again. could be
            ## that another process just stared up and created
            ## the pid file but hadn't yet written it's pid down.
            sleep 3;

            return $self->lock(++$step);
        }
    }

    ## if we got this far something went wrong.
    ## let's try again.
    return $self->lock(++$step);

}


=head2 unlock

Remove PID file.

=cut

sub unlock
{
    my $self = shift;

    ## check if we have open file handle on pid.
    if (my $fh = $self->{_fh}) {

        ## well okay then, let's close it and remove the file.
        $fh->close;
        ## remove the pid file.
        unless (unlink $self->{file}) {
            ## that's strange, couldn't remove it. let's just warn tho.
            $self->{error} = "couldn't remove pid file for some reason: $!";
            carp $self->{error} if $^W;
            return undef;
        }
        undef $self->{_fh};
        return 1;
    }

    return 0;

}

sub DESTROY
{
    ## when we go out of scope, be sure to remove the pid file.
    $_[0]->unlock;
}


1;
